import os
from typing import List, Dict, Tuple, Any
from dotenv import load_dotenv
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession, AsyncEngine
from sqlalchemy.orm import sessionmaker
from Exception.ControlledException import DatabaseException
from Utils.Code import DB_SUCCESS
from Utils.Message import DB_STRING_ERROR

load_dotenv()

class Connection:

    def __init__(self, connection_string: str) -> None:
        if not connection_string:
            raise ValueError(DB_STRING_ERROR)
        
        self.connection_string = connection_string
        self.engine = None
        
    async def get_engine(self) -> AsyncEngine:
        if not self.engine:
            self.engine = create_async_engine(
                self.connection_string, 
                pool_size=15,  
                max_overflow=10, 
                pool_timeout=30,
                pool_recycle=1800,
                echo=os.getenv('DB_DEBUG', 'False').lower() == 'true', 
                pool_pre_ping=True
            )

        return self.engine

    async def get_session(self) -> AsyncSession:
        self.engine = await self.get_engine()
        session_factory = sessionmaker(
            bind=self.engine,
            class_=AsyncSession,
            expire_on_commit=False
        )
        return session_factory()

    async def execute_query_return_first_value(self, query: str, params: Dict[str, Any] = None) -> Any:
        async with (await self.get_session()) as session:
            async with session.begin():
                result = await session.execute(text(query), params)
                return result.scalar()

    async def execute_query_return_data(self, query: str, params: Dict[str, Any] = None) -> Dict[str, Any]:
        keys, row = await self._execute_query(query=query, params=params, fetchone=True)
        if not row:
            raise DatabaseException(error=row.STATUS_MESSAGE, status_code=row.STATUS_CODE)
        
        return dict(zip(keys, row))

    async def execute_query_return_list(self, query: str, params: Dict[str, Any] = None) -> List[Dict[str, Any]]:
        keys, rows = await self._execute_query(query=query, params=params)
        if not rows:
            return []
        
        return [dict(zip(keys, row)) for row in rows]

    async def execute_query(self, query: str, params: Dict[str, Any] = None, code:str = DB_SUCCESS) -> str:
        async with (await self.get_session()) as session:
            async with session.begin():
                result = await session.execute(text(query), params)
                row = result.fetchone()
                if row.STATUS_CODE != code:
                    raise DatabaseException(error=row.STATUS_MESSAGE, status_code=row.STATUS_CODE)
                
                return row.STATUS_MESSAGE
            
    
    async def _execute_query(self, query:str, params:Dict[str, Any], fetchone=False) -> Tuple[Any, Any]:
        async with (await self.get_session()) as session:
            async with session.begin():
                result = await session.execute(text(query), params)
                columns_name = result.keys()
                if fetchone:
                    return columns_name, result.fetchone()
                
                return columns_name, result.fetchall()


    async def close_engine(self) -> None:
        await self.engine.dispose()